/*
  Delta filesystem
  Copyright (C) 2009  Miklos Szeredi <miklos@szeredi.hu>

  This program can be distributed under the terms of the GNU GPLv2.

  gcc -Wall `pkg-config fuse --cflags --libs` deltafs.c -o deltafs

  todo:
     - optimize directory merging
     - tempfiles in tempdir
*/

#define FUSE_USE_VERSION 26
#define _REENTRANT
#define _GNU_SOURCE

#ifndef FUSE_SET_ATTR_ATIME_NOW
#define FUSE_SET_ATTR_ATIME_NOW (1 << 7)
#endif
#ifndef FUSE_SET_ATTR_MTIME_NOW
#define FUSE_SET_ATTR_MTIME_NOW (1 << 8)
#endif

#include "config.h"
#include <fuse_lowlevel.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <err.h>
#include <assert.h>

struct deltafs_config {
	char *lowerdir;
	char *upperdir;
	int o_direct;
};

struct deltafs_config deltafs;
static int lowerfd;
static int upperfd;
static int metafd;
static const char *overlay_name = "overlay";
static const char *mergedir_name = "merge";
static const char *whiteout_name = "WHITEOUT";
static const char *opaque_name = "OPAQUE";
static bool whiteout_created = false;
static bool opaque_created = false;
static ino_t whiteout_ino;
static ino_t opaque_ino;
static double attr_timeout = 999999999.0;
static double entry_timeout = 999999999.0;
static double negative_timeout = 999999999.0;
static unsigned long long inoctr = 2;
typedef char pathstring[PATH_MAX + 1];
static unsigned pagesize;
static struct fuse_chan *fuse_ch;

#define MODEMASK 07777

enum deltafs_type {
	DF_UNKNOWN,
	DF_LOWER,
	DF_MERGE,
	DF_UPPER,
	DF_DELETED,
	DF__MERGEDIR,
};

struct deltafs_inode {
	fuse_ino_t ino;
	uint64_t nlookup;
	struct deltafs_inode *id_next;
	struct deltafs_entry *alias;
};

struct deltafs_entry {
	char *name;
	enum deltafs_type type;
	struct deltafs_inode *inode;
	struct deltafs_entry *name_next;
	struct deltafs_entry **name_prevp;
	struct deltafs_entry *alias_next;
	struct deltafs_entry *parent;
};

struct lookup_result {
	int fd;
	pathstring path;
	bool closefd;
};

struct deltafs_inode **id_table;
static size_t id_table_size;
struct deltafs_entry **name_table;
static size_t name_table_size;

static uint64_t twang64_hash(uint64_t key)
{
	key = (~key) + (key << 21); // key = (key << 21) - key - 1;
	key = key ^ (key >> 24);
	key = (key + (key << 3)) + (key << 8); // key * 265
	key = key ^ (key >> 14);
	key = (key + (key << 2)) + (key << 4); // key * 21
	key = key ^ (key >> 28);
	key = key + (key << 31);

	return key;
}

static size_t id_hash(fuse_ino_t ino)
{
	return twang64_hash(ino) % id_table_size;
}

static size_t name_hash(fuse_ino_t parent, const char *name)
{
	uint64_t hash = parent;

	for (; *name; name++)
		hash = hash * 31 + (unsigned char) *name;

	return twang64_hash(hash) % name_table_size;
}

static struct deltafs_inode *get_inode(fuse_ino_t ino)
{
	struct deltafs_inode *inode = id_table[id_hash(ino)];

	while (inode && inode->ino != ino)
		inode = inode->id_next;

	if (!inode)
		errx(1, "internal error: inode %li not found", ino);

	return inode;
}


static struct deltafs_entry *get_entry(fuse_ino_t ino)
{
	return get_inode(ino)->alias;
}

static bool entry_has_alias(struct deltafs_entry *entry)
{
	struct deltafs_inode *inode = entry->inode;

	return inode->alias != entry || entry->alias_next != NULL;
}

static void remove_alias(struct deltafs_entry *entry)
{
	struct deltafs_inode *inode = entry->inode;
	struct deltafs_entry **ep;

	for (ep = &inode->alias; *ep != NULL; ep = &(*ep)->alias_next) {
		if (*ep == entry) {
			*ep = entry->alias_next;
			entry->alias_next = NULL;
			return;
		}
	}

	warnx("internal error: entry %s/%lli not found in aliases",
	      entry->name, (unsigned long long) entry->parent->inode->ino);
}

static void add_to_id_hash(struct deltafs_inode *inode)
{
	size_t idh = id_hash(inode->ino);

	inode->id_next = id_table[idh];
	id_table[idh] = inode;
}

static void free_inode(struct deltafs_inode *inode)
{
	free(inode);
}

static void free_entry(struct deltafs_entry *entry)
{
	free(entry->name);
	free(entry);
}

static struct deltafs_inode *create_inode(fuse_ino_t ino)
{
	struct deltafs_inode *inode;

	inode = malloc(sizeof(struct deltafs_inode));
	if (inode) {
		memset(inode, 0, sizeof(struct deltafs_inode));
		inode->ino = ino;
		inode->nlookup = 0;
	}
	return inode;
}

static struct deltafs_entry *create_entry(struct deltafs_entry *parent,
					  const char *name,
					  struct deltafs_inode *inode,
					  enum deltafs_type type)
{
	struct deltafs_entry *entry;

	entry = malloc(sizeof(struct deltafs_entry));
	if (entry) {
		memset(entry, 0, sizeof(struct deltafs_entry));
		entry->name = strdup(name);
		if (!entry->name) {
			free(entry);
			entry = NULL;
		} else {
			entry->parent = parent;
			entry->inode = inode;
			entry->type = type;
			if (inode != NULL) {
				entry->alias_next = inode->alias;
				inode->alias = entry;
			}
		}
	}

	return entry;
}

static void add_to_name_hash(struct deltafs_entry *entry)
{
	size_t hash = name_hash(entry->parent->inode->ino, entry->name);
	struct deltafs_entry **ep = &name_table[hash];
	struct deltafs_entry *first = *ep;

	entry->name_next = first;
	if (first)
		first->name_prevp = &entry->name_next;
	*ep = entry;
	entry->name_prevp = ep;
}

static void remove_from_name_hash(struct deltafs_entry *entry)
{
	if (entry->name_prevp) {
		struct deltafs_entry *next = entry->name_next;
		struct deltafs_entry **prevp = entry->name_prevp;

		*prevp = next;
		if (next)
			next->name_prevp = prevp;

		entry->name_prevp = NULL;
		entry->name_next = NULL;
	}
}

static bool is_name_unhashed(struct deltafs_entry *entry)
{
	return entry->name_prevp == NULL;
}

static bool is_deleted(struct deltafs_entry *entry)
{
	return is_name_unhashed(entry) && entry->inode->ino != FUSE_ROOT_ID &&
		entry->inode != NULL;
}

static char *add_name(char *path, char *s, const char *name, int slash)
{
	size_t namelen = strlen(name);

	if (s - path < (long) namelen + (slash ? 1 : 0))
		return NULL;

	if (slash) {
		s--;
		*s = '/';
	}
	s -= namelen;
	memcpy(s, name, namelen);

	return s;
}

static void lookup_finish(struct lookup_result *lr)
{
	if (lr->closefd)
		close(lr->fd);
}

static const char *entry_name(enum deltafs_type type,
			      struct deltafs_entry *entry)
{
	if (entry->inode && entry->inode->ino == FUSE_ROOT_ID) {
		switch (type) {
		case DF_UPPER:
		case DF_MERGE:
			return overlay_name;
		case DF__MERGEDIR:
			return mergedir_name;
		default:;
		}
	}
	return entry->name;
}

static int lookup_path(enum deltafs_type type, struct lookup_result *lr,
		       struct deltafs_entry *entry)
{
	int res;
	char *s = lr->path + sizeof(pathstring) - 1;

	lr->closefd = false;

	switch (type) {
	case DF_UPPER:
	case DF_MERGE:
	case DF__MERGEDIR:
		lr->fd = upperfd;
		break;

	case DF_LOWER:
		lr->fd = lowerfd;
		break;

	case DF_DELETED:
		lr->fd = metafd;
		res = snprintf(lr->path, sizeof(pathstring), "deleted.%llu",
			       (unsigned long long) entry->inode->ino);
		if (res < 0 || res >= (int) sizeof(pathstring))
			return -ENAMETOOLONG;

		return 0;

	default:
		errx(1, "internal error: invalid type %i in lookup_path", type);
	}
	*s = '\0';
	s = add_name(lr->path, s, entry_name(type, entry), 0);
	if (!s)
		return -ENAMETOOLONG;

	while (entry->parent != NULL) {
		char *ns;

		entry = entry->parent;
		if (entry->inode->ino == FUSE_ROOT_ID && type == DF_LOWER)
			break;

		ns = add_name(lr->path, s, entry_name(type, entry), 1);
		if (!ns) {
			struct lookup_result lr2;
			int res;
			int newfd;

			res = lookup_path(type, &lr2, entry);
			if (res < 0)
				return res;

			newfd = openat(lr2.fd, lr2.path,
				       O_RDONLY | O_DIRECTORY | O_NOFOLLOW);
			res = -errno;
			lookup_finish(&lr2);
			if (newfd == -1)
				return res;

			lr->fd = newfd;
			lr->closefd = true;
			break;
		}
		s = ns;
	}
	memmove(lr->path, s, sizeof(pathstring) - (s - lr->path));

	return 0;
}

static int open_ino(enum deltafs_type type, struct deltafs_entry *entry,
		    int flags)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = openat(lr.fd, lr.path, flags);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int stat_ino(enum deltafs_type type, struct deltafs_entry *entry,
		    struct stat *st)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = fstatat(lr.fd, lr.path, st, AT_SYMLINK_NOFOLLOW);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int unlink_ino(enum deltafs_type type, struct deltafs_entry *entry,
		      int flags)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = unlinkat(lr.fd, lr.path, flags);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int link_to_ino(int fd, const char *name,
		       enum deltafs_type type, struct deltafs_entry *entry)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = linkat(fd, name, lr.fd, lr.path, 0);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int link_from_ino(enum deltafs_type type, struct deltafs_entry *entry,
			 int fd, const char *name)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = linkat(lr.fd, lr.path, fd, name, 0);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int mkdir_ino(enum deltafs_type type, struct deltafs_entry *entry,
		     mode_t mode)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = mkdirat(lr.fd, lr.path, mode);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int readlink_ino(enum deltafs_type type, struct deltafs_entry *entry,
			char **link)
{
	struct lookup_result lr;
	int size = 4096;
	char *buf;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	do {
		buf = malloc(size);
		if (buf == NULL) {
			res = -ENOMEM;
			break;
		}
		res = readlinkat(lr.fd, lr.path, buf, size);
		if (res == -1) {
			res = -errno;
			free(buf);
			break;
		}
		if (res < size) {
			buf[res] = '\0';
			*link = buf;
			res = 0;
			break;
		}
		free(buf);
		size *= 2;
	} while (true);

	lookup_finish(&lr);

	return res;
}

static bool is_whiteout(ino_t ino)
{
	return whiteout_created && ino == whiteout_ino;
}

static bool is_opaque(struct deltafs_entry *entry)
{
	if (opaque_created) {
		struct stat st;
		int res;

		res = stat_ino(DF__MERGEDIR, entry, &st);
		if (res == 0 && st.st_ino == opaque_ino)
			return true;
	}

	return false;
}

static int check_type(struct deltafs_entry *entry, enum deltafs_type *type,
		      struct stat *st)
{
	int res;

	res = stat_ino(DF_UPPER, entry, st);
	if (!res) {
		if (!S_ISDIR(st->st_mode)) {
			if (is_whiteout(st->st_ino))
				return -ENOENT;

			*type = DF_UPPER;
		} else {
			struct stat xst;

			res = stat_ino(DF_LOWER, entry, &xst);
			if (res < 0) {
				if (res == -ENOENT) {
					*type = DF_UPPER;
					res = 0;
				}
			} else if (S_ISDIR(xst.st_mode)) {
				if (is_opaque(entry))
					*type = DF_UPPER;
				else
					*type = DF_MERGE;
			} else {
				*type = DF_UPPER;
			}
		}
	} else if (res == -ENOENT) {
		res = stat_ino(DF_LOWER, entry, st);
		if (!res)
			*type = DF_LOWER;
	}

	return res;
}

static struct deltafs_entry *find_entry(fuse_ino_t parent, const char *name)
{
	struct deltafs_entry *entry = name_table[name_hash(parent, name)];

	while (entry) {
		if (entry->parent && entry->parent->inode->ino == parent &&
		    strcmp(entry->name, name) == 0)
			break;

		entry = entry->name_next;
	}

	return entry;
}

static struct deltafs_entry *new_entry(fuse_ino_t parent, const char *name)
{
	struct deltafs_entry *entry;
	struct deltafs_inode *inode;

	inode = create_inode(inoctr++);
	if (!inode)
		return NULL;

	entry = create_entry(get_entry(parent), name, inode, DF_UNKNOWN);
	if (!entry) {
		free_inode(inode);
		return NULL;
	}
	add_to_id_hash(inode);
	add_to_name_hash(entry);

	return entry;
}

static int do_lookup(fuse_ino_t parent, const char *name,
		     struct fuse_entry_param *e)
{
	int res;
	struct deltafs_entry *entry = find_entry(parent, name);

	if (!entry) {
		entry = new_entry(parent, name);
		if (!entry)
			return -ENOMEM;
	}

	if (entry->type == DF_UNKNOWN && entry->parent->type != DF_MERGE)
		entry->type = entry->parent->type;

	if (entry->type == DF_UNKNOWN)
		res = check_type(entry, &entry->type, &e->attr);
	else
		res = stat_ino(entry->type, entry, &e->attr);

	if (res == -ENOENT) {
		remove_from_name_hash(entry);
		entry->type = DF_UNKNOWN;
		/* remove entry */
		e->ino = 0;
		e->entry_timeout = negative_timeout;
		return 0;
	}

	if (res < 0)
		return res;

	entry->inode->nlookup++;

	e->ino = entry->inode->ino;
	e->attr.st_ino = e->ino;
	e->generation = 0;
	e->attr_timeout = attr_timeout;
	e->entry_timeout = entry_timeout;

	return 0;
}

static int set_owner(int rfd, const char *path, struct stat *st)
{
	int res;

	res = fchownat(rfd, path, st->st_uid, st->st_gid,
		       AT_SYMLINK_NOFOLLOW);
	if (res == -1)
		return -errno;

	return 0;
}

static int set_times(int rfd, const char *path, struct stat *st)
{
	int res;
	struct timespec times[2];

	times[0] = st->st_atim;
	times[1] = st->st_mtim;
	res = utimensat(rfd, path, times, AT_SYMLINK_NOFOLLOW);
	if (res == -1)
		return -errno;

	return 0;
}

static int set_times_ino(enum deltafs_type type, struct deltafs_entry *entry,
			 struct stat *st)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = set_times(lr.fd, lr.path, st);
	lookup_finish(&lr);

	return res;
}

static int setattr_path(int rfd, const char *path, struct stat *st, int to_set)
{
	int res;

	if (to_set & FUSE_SET_ATTR_MODE) {
		res = fchmodat(rfd, path, st->st_mode, 0);
		if (res == -1)
			return -errno;
	}
	if (to_set & (FUSE_SET_ATTR_UID | FUSE_SET_ATTR_GID)) {
		uid_t uid = (to_set & FUSE_SET_ATTR_UID) ?
			st->st_uid : (uid_t) -1;
		gid_t gid = (to_set & FUSE_SET_ATTR_GID) ?
			st->st_gid : (gid_t) -1;

		res = fchownat(rfd, path, uid, gid, AT_SYMLINK_NOFOLLOW);
		if (res == -1)
			return -errno;
	}
	if (to_set & FUSE_SET_ATTR_SIZE) {
		int fd;

		fd = openat(rfd, path, O_WRONLY);
		if (fd == -1)
			return -errno;

		res = ftruncate(fd, st->st_size);
		if (res == -1) {
			res = -errno;
			close(fd);
			return res;
		}
		close(fd);
	}
	if (to_set & (FUSE_SET_ATTR_ATIME | FUSE_SET_ATTR_MTIME)) {
		struct timespec times[2];

		times[0].tv_sec = 0;
		times[1].tv_sec = 0;
		times[0].tv_nsec = UTIME_OMIT;
		times[1].tv_nsec = UTIME_OMIT;

		if (to_set & FUSE_SET_ATTR_ATIME_NOW)
			times[0].tv_nsec = UTIME_NOW;
		else if (to_set & FUSE_SET_ATTR_ATIME)
			times[0] = st->st_atim;

		if (to_set & FUSE_SET_ATTR_MTIME_NOW)
			times[1].tv_nsec = UTIME_NOW;
		else if (to_set & FUSE_SET_ATTR_MTIME)
			times[1] = st->st_mtim;

		res = utimensat(rfd, path, times, AT_SYMLINK_NOFOLLOW);
		if (res == -1)
			return -errno;
	}
	return 0;
}

static int setattr_ino(enum deltafs_type type, struct deltafs_entry *entry,
		       struct stat *st, int to_set)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = setattr_path(lr.fd, lr.path, st, to_set);
	lookup_finish(&lr);

	return res;
}

/* fixme: file holes */
static int copy_data(int src, int dst, off_t size)
{
	off_t to_copy = size;
	int bufsize = 128 * 1024;
	void *buf;
	int res;

	res = posix_memalign(&buf, pagesize, bufsize);
	if (res != 0)
		return -res;

	while (to_copy > 0) {
		int num = bufsize;

		if (to_copy < num)
			num = (to_copy + pagesize - 1) & ~(pagesize - 1);

		res = read(src, buf, num);
		if (res == -1) {
			res = -errno;
			break;
		}
		if (res == 0) {
			res = ftruncate(dst, size);
			if (res == -1)
				res = -errno;
			break;
		}

		num = to_copy < res ? to_copy : res;
		res = write(dst, buf, num);
		if (res == -1) {
			res = -errno;
			break;
		}
		if (num != res) {
			res = -EIO;
			break;
		}
		to_copy -= num;
		res = 0;
	}
	free(buf);

	return res;
}

static int rename_to_ino(int rfd, const char *name, enum deltafs_type type,
			 struct deltafs_entry *entry)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(type, &lr, entry);
	if (res < 0)
		return res;

	res = renameat(rfd, name, lr.fd, lr.path);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int rename_from_entry(struct deltafs_entry *entry,
			     int rfd, const char *name)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(DF_UPPER, &lr, entry);
	if (res < 0)
		return res;

	res = renameat(lr.fd, lr.path, rfd, name);
	if (res == -1)
		res = -errno;
	lookup_finish(&lr);

	return res;
}

static int create_opaque_node(void)
{
	int res;
	struct stat st;

	if (opaque_created)
		return 0;

	res = symlinkat("(OPAQUE)", metafd, opaque_name);
	if (res == -1)
		return -errno;

	res = fstatat(metafd, opaque_name, &st, AT_SYMLINK_NOFOLLOW);
	if (res == -1)
		return -errno;

	opaque_created = true;
	opaque_ino = st.st_ino;

	return 0;
}

static int create_opaque_tree(struct deltafs_entry *entry)
{
	int res;
	struct stat st;

	res = stat_ino(DF__MERGEDIR, entry, &st);
	if (res == 0) {
		if (!S_ISDIR(st.st_mode)) {
			warnx("invalid opaque tree");
			return -EIO;
		}
		return 0;
	}

	if (entry->parent) {
		res = create_opaque_tree(entry->parent);
		if (res < 0)
			return res;
	}

	return mkdir_ino(DF__MERGEDIR, entry, 0700);
}

static int create_opaque_entry(struct deltafs_entry *entry)
{
	int res;

	res = create_opaque_node();
	if (res < 0)
		return res;

	res = create_opaque_tree(entry->parent);
	if (res < 0)
		return res;

	return link_to_ino(metafd, opaque_name, DF__MERGEDIR, entry);
}

static int rename_dir_to_whiteout(int rfd, const char *name,
				  struct deltafs_entry *entry)
{
	struct stat st;
	int res;
	bool opaque = false;

	res = stat_ino(DF_UPPER, entry, &st);
	if (res < 0)
		return res;

	if (!is_whiteout(st.st_ino))
		return -ENOTDIR;

	res = stat_ino(DF_LOWER, entry, &st);
	if (res < 0 && res != -ENOENT)
		return res;

	if (res == 0 && S_ISDIR(st.st_mode)) {
		res = create_opaque_entry(entry);
		if (res < 0)
			return res;

		opaque = true;
	}

	unlink_ino(DF_UPPER, entry, 0);

	res = rename_to_ino(rfd, name, DF_UPPER, entry);
	if (res < 0 && opaque)
		unlink_ino(DF__MERGEDIR, entry, 0);

	return res;
}

static int copy_up_regular(int tmpfd, const char *tmpfile,
			   struct deltafs_entry *entry, struct stat *st)
{
	int res = 0;
	int src;
	int dst;

	dst = openat(tmpfd, tmpfile, O_WRONLY | O_CREAT | O_EXCL, st->st_mode);
	if (dst == -1)
		return -errno;

	if (st->st_size != 0) {
		int openflags = O_RDONLY | O_NOFOLLOW;

		if (deltafs.o_direct)
			openflags |= O_DIRECT;

		src = open_ino(DF_LOWER, entry, openflags);
		if (src < 0) {
			res = src;
		} else {
			res = copy_data(src, dst, st->st_size);
			close(src);
		}
	}
	close(dst);

	return res;
}

static int copy_up(struct deltafs_entry *entry, struct stat *st)
{
	int res;
	int tmpfd = metafd;
	char *tmpfile = "tmp.copy";
	struct stat parst;
	struct deltafs_entry *parent = NULL;
	enum deltafs_type newtype;
	int unlinkflag = 0;

	if (!is_name_unhashed(entry))
		parent = entry->parent;

	if (parent) {
		res = stat_ino(DF_LOWER, parent, &parst);
		if (res < 0)
			return res;

		assert(parent->type == DF_LOWER || parent->type == DF_MERGE);
		if (parent->type == DF_LOWER) {
			res = copy_up(parent, &parst);
			if (res < 0)
				return res;
		}
	}

	if (S_ISDIR(st->st_mode)) {
		unlinkflag |= AT_REMOVEDIR;
		res = mkdirat(tmpfd, tmpfile, st->st_mode);
		if (res == -1)
			res = -errno;
	} else if (S_ISREG(st->st_mode)) {
		res = copy_up_regular(tmpfd, tmpfile, entry, st);
	} else if (S_ISLNK(st->st_mode)) {
		char *linkbuf;

		res = readlink_ino(DF_LOWER, entry, &linkbuf);
		if (res == 0) {
			res = symlinkat(linkbuf, tmpfd, tmpfile);
			if (res == -1)
				res = -errno;
			free(linkbuf);
		}
	} else {
		res = mknodat(tmpfd, tmpfile, st->st_mode, st->st_rdev);
		if (res == -1)
			res = -errno;
	}

	if (res == 0)
		res = set_owner(tmpfd, tmpfile, st);
	if (res == 0 && (st->st_atim.tv_nsec != UTIME_NOW ||
			 st->st_mtim.tv_nsec != UTIME_NOW)) {
		res = set_times(tmpfd, tmpfile, st);
	}
	if (res == 0 && (S_ISREG(st->st_mode) || S_ISDIR(st->st_mode))) {
		int fd = openat(tmpfd, tmpfile, O_RDONLY | O_NOFOLLOW);
		if (fd == -1)
			res = -errno;
		else {
			res = fsync(fd);
			if (res == -1)
				res = -errno;

			close(fd);
		}
	}
	if (is_deleted(entry)) {
		newtype = DF_DELETED;
	} else {
		if (S_ISDIR(st->st_mode))
			newtype = DF_MERGE;
		else
			newtype = DF_UPPER;
	}

	if (res == 0)
		res = rename_to_ino(tmpfd, tmpfile, newtype, entry);

	if (res == 0) {
		/*
		 * restoring times on parent is best effort, we don't
		 * error out if it fails for some reason (but it
		 * shouldn't normally fail)
		 */
		if (parent)
			set_times_ino(DF_UPPER, parent, &parst);

		entry->type = newtype;
	} else {
		unlinkat(tmpfd, tmpfile, unlinkflag);
	}

	return res;
}

static int do_getattr(fuse_ino_t ino, struct stat *st)
{
	struct deltafs_entry *entry = get_entry(ino);
	int res;

	res = stat_ino(entry->type, entry, st);
	if (res < 0)
		return res;

	if (is_deleted(entry)) {
		st->st_nlink = 0;
		if (entry->type == DF_LOWER) {
			struct stat whst;

			res = stat_ino(DF_UPPER, entry, &whst);
			if (res == 0)
				st->st_ctime = whst.st_ctime;
		}
	}
	st->st_ino = entry->inode->ino;

	return 0;
}

struct delta_dirent {
	mode_t mode;
	ino_t ino;
	char *name;
	struct delta_dirent *next;
};

struct delta_dirp {
	DIR *dp;
	bool read;
	struct dirent *entry;
	off_t offset;
	struct delta_dirent *first;
};

static int entry_to_stat(struct deltafs_entry *dir, struct dirent *entry,
			  struct stat *st)
{
	const char *name;

	name = entry->d_name;
	memset(st, 0, sizeof(struct stat));
	st->st_mode = entry->d_type << 12;
	st->st_ino = entry->d_ino;
	if (strcmp(name, ".") == 0) {
		st->st_ino = dir->inode->ino;
	} else if (strcmp(name, "..") == 0) {
		if (dir->inode->ino == FUSE_ROOT_ID)
			st->st_ino = FUSE_ROOT_ID;
		else
			st->st_ino = dir->parent->inode->ino;
	} else  {
		struct deltafs_entry *entry2;

		entry2 = find_entry(dir->inode->ino, name);
		if (!entry2) {
			entry2 = new_entry(dir->inode->ino, name);
			if (entry2 == NULL)
				return -ENOMEM;
		}

		st->st_ino = entry2->inode->ino;
	}

	return 0;
}

static int my_readdir(DIR *dp, struct dirent **entp)
{
	errno = 0;
	*entp = readdir(dp);
	if (*entp == NULL) {
		if (errno)
			return -errno;
		return 0;
	}
	return 1;
}

static int merge_dir(struct deltafs_entry *dir, struct delta_dirent **listp,
		     DIR *dp, bool check_whiteout)
{
	struct delta_dirent *dent;
	struct delta_dirent **dentp;
	struct dirent *entry;
	struct stat st;
	int res;

	while (1) {
		res = my_readdir(dp, &entry);
		if (res <= 0)
			return res;

		for (dentp = listp; *dentp; dentp = &(*dentp)->next) {
			dent = *dentp;
			if (strcmp(dent->name, entry->d_name) == 0)
				break;
		}
		if (*dentp != NULL)
			continue;

		res = entry_to_stat(dir, entry, &st);
		if (res < 0)
			return res;
		dent = malloc(sizeof(struct delta_dirent));
		if (dent == NULL)
			return -ENOMEM;
		dent->name = strdup(entry->d_name);
		if (dent->name == NULL) {
			free(dent);
			return -ENOMEM;
		}

		dent->mode = st.st_mode;
		dent->ino = st.st_ino;
		dent->next = NULL;

		if (check_whiteout && is_whiteout(entry->d_ino))
			dent->ino = 0;

		*dentp = dent;
	}
}

static int opendir_ino(enum deltafs_type type, struct deltafs_entry *entry,
		       DIR **dp)
{
	int res;
	int fd;

	fd = open_ino(type, entry, O_RDONLY | O_DIRECTORY | O_NOFOLLOW);
	if (fd < 0)
		return fd;

	*dp = fdopendir(fd);
	if (!*dp) {
		res = -errno;
		close(fd);
		return res;
	}

	return 0;
}

static void free_dirlist(struct delta_dirent **listp)
{
	while (*listp) {
		struct delta_dirent *e = *listp;

		*listp = e->next;
		free(e->name);
		free(e);
	}
}

static int merge_directories(struct deltafs_entry *dir,
			     struct delta_dirent **listp)
{
	int res;
	DIR *udp;
	DIR *ldp;

	res = opendir_ino(DF_UPPER, dir, &udp);
	if (res < 0)
		return res;

	res = opendir_ino(DF_LOWER, dir, &ldp);
	if (res < 0) {
		closedir(udp);
		return res;
	}

	res = merge_dir(dir, listp, udp, true);
	if (res == 0)
		res = merge_dir(dir, listp, ldp, false);
	closedir(udp);
	closedir(ldp);

	if (res < 0)
		free_dirlist(listp);

	return res;
}

static int do_opendir(fuse_ino_t ino, struct delta_dirp *d)
{
	struct deltafs_entry *dir = get_entry(ino);
	int res;

	memset(d, 0, sizeof(struct delta_dirp));

	if (dir->type == DF_MERGE) {
		d->dp = NULL;
		d->read = false;
		return 0;
	}

	res = opendir_ino(dir->type, dir, &d->dp);
	if (res < 0)
		return res;

	d->offset = 0;
	d->entry = NULL;

	return 0;
}

static int real_readdir(fuse_req_t req, fuse_ino_t ino, char *buf, size_t size,
			off_t off, struct delta_dirp *d)
{
	size_t bufpos = 0;
	struct deltafs_entry *entry = get_entry(ino);
	int res;

	if (off != d->offset) {
		seekdir(d->dp, off);
		d->entry = NULL;
		d->offset = off;
	}
	while (size - bufpos) {
		struct stat st;
		off_t nextoff;
		size_t entsize;

		if (!d->entry) {
			res = my_readdir(d->dp, &d->entry);
			if (res < 0)
				return res;
			if (!res)
				break;
		}
		res = entry_to_stat(entry, d->entry, &st);
		if (res < 0)
			return res;
		nextoff = telldir(d->dp);
		entsize = fuse_add_direntry(req, buf + bufpos, size - bufpos,
					    d->entry->d_name, &st, nextoff);
		if (entsize > size - bufpos)
			break;

		bufpos += entsize;
		d->entry = NULL;
		d->offset = nextoff;
	}
	return bufpos;
}

static int buffer_readdir(fuse_req_t req, fuse_ino_t ino, char *buf,
			  size_t size, off_t off, struct delta_dirp *d)
{
	size_t bufpos = 0;
	struct delta_dirent *dent;
	off_t cnt;

	if (!d->read) {
		struct deltafs_entry *dir = get_entry(ino);
		int res;

		res = merge_directories(dir, &d->first);
		if (res < 0)
			return res;

		d->read = true;
	}

	dent = d->first;
	for (cnt = 0; cnt < off && dent; cnt++)
		dent = dent->next;

	d->offset = off;

	while (dent && size - bufpos) {
		struct stat st;
		off_t nextoff = d->offset + 1;
		struct delta_dirent *nextdent = dent->next;
		size_t entsize;

		while (nextdent) {
			if (nextdent->ino)
				break;
			nextoff++;
			nextdent = nextdent->next;
		}

		memset(&st, 0, sizeof(struct stat));
		st.st_mode = dent->mode;
		st.st_ino = dent->ino;

		entsize = fuse_add_direntry(req, buf + bufpos, size - bufpos,
					    dent->name, &st, nextoff);
		if (entsize > size - bufpos)
			break;

		bufpos += entsize;
		d->offset = nextoff;
		dent = nextdent;
	}
	return bufpos;
}

static int do_readdir(fuse_req_t req, fuse_ino_t ino, char *buf, size_t size,
		      off_t off, struct delta_dirp *d)
{
	if (d->dp)
		return real_readdir(req, ino, buf, size, off, d);
	else
		return buffer_readdir(req, ino, buf, size, off, d);
}

static void delta_lookup(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	struct fuse_entry_param e;
	int res = do_lookup(parent, name, &e);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_entry(req, &e);
}

static int cleanup_dir(struct deltafs_entry *entry)
{
	int res;
	int fd;
	DIR *dp;
	struct dirent *de;

	res = opendir_ino(entry->type, entry, &dp);
	if (res < 0)
		return res;

	fd = dirfd(dp);

	res = 0;
	while ((de = readdir(dp)) != NULL) {
		if (strcmp(de->d_name, ".") == 0 ||
		    strcmp(de->d_name, "..") == 0)
			continue;

		res = unlinkat(fd, de->d_name, 0);
		if (res == -1) {
			res = -errno;
			break;
		}
	}
	closedir(dp);
	if (res == 0)
		res = unlink_ino(entry->type, entry, AT_REMOVEDIR);

	return res;
}

static void clean_up_deleted(struct deltafs_entry *entry)
{
	int res = unlink_ino(entry->type, entry, 0);
	if (res == -EISDIR) {
		res = unlink_ino(entry->type, entry, AT_REMOVEDIR);
		if (res == -ENOTEMPTY)
			res = cleanup_dir(entry);
	}

	if (res < 0) {
		warnx("failed to remove deleted.%llu: %s",
		      (unsigned long long) entry->inode->ino, strerror(-res));
	}
}

static void delta_forget(fuse_req_t req, fuse_ino_t ino, unsigned long nlookup)
{
	struct deltafs_inode *inode = get_inode(ino);
	if (inode->nlookup < nlookup) {
		errx(1, "internal error: forget ino=%llu %llu from %llu",
		     (unsigned long long) ino, (unsigned long long) nlookup,
		     (unsigned long long) inode->nlookup);
	}

	inode->nlookup -= nlookup;
	if (inode->nlookup == 0) {
		struct deltafs_entry *entry = inode->alias;

		if (entry->type == DF_DELETED)
			clean_up_deleted(entry);


		/* remove entry */
	}
	fuse_reply_none(req);
}

/* fixme: st_nlink on merged directories is wrong */
static void delta_getattr(fuse_req_t req, fuse_ino_t ino,
			  struct fuse_file_info *fi)
{
	struct stat st;
	int res;

	(void) fi;
	res = do_getattr(ino, &st);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_attr(req, &st, attr_timeout);
}

static bool times_eq(struct timespec *tv1, struct timespec *tv2)
{
	return tv1->tv_sec == tv2->tv_sec && tv1->tv_nsec == tv2->tv_nsec;
}

static bool copy_attrs(struct stat *oldst, struct stat *newst, int to_set)
{
	bool changed = false;

	newst->st_mode &= MODEMASK;

	if ((to_set & FUSE_SET_ATTR_MODE) &&
	    (oldst->st_mode & MODEMASK) != newst->st_mode) {
		changed = true;
		oldst->st_mode = (oldst->st_mode & S_IFMT) | newst->st_mode;
	}
	if ((to_set & FUSE_SET_ATTR_UID) && oldst->st_uid != newst->st_uid) {
		changed = true;
		oldst->st_uid = newst->st_uid;
	}
	if ((to_set & FUSE_SET_ATTR_GID) && oldst->st_gid != newst->st_gid) {
		changed = true;
		oldst->st_gid = newst->st_gid;
	}
	if ((to_set & FUSE_SET_ATTR_SIZE) && oldst->st_size != newst->st_size) {
		changed = true;
		oldst->st_size = newst->st_size;
	}
	if (to_set & FUSE_SET_ATTR_ATIME_NOW) {
		changed = true;
		oldst->st_atim.tv_sec = 0;
		oldst->st_atim.tv_nsec = UTIME_NOW;
	} else if ((to_set & FUSE_SET_ATTR_ATIME) && 
		   !times_eq(&oldst->st_atim, &newst->st_atim)) {
		changed = true;
		oldst->st_atim = newst->st_atim;
	}
	if (to_set & FUSE_SET_ATTR_MTIME_NOW) {
		changed = true;
		oldst->st_mtim.tv_sec = 0;
		oldst->st_mtim.tv_nsec = UTIME_NOW;
	} else if ((to_set & FUSE_SET_ATTR_MTIME) && 
		   !times_eq(&oldst->st_mtim, &newst->st_mtim)) {
		changed = true;
		oldst->st_mtim = newst->st_mtim;
	}

	return changed;
}

static int do_setattr(fuse_ino_t ino, struct stat *st, int to_set)
{
	struct deltafs_entry *entry = get_entry(ino);
	int res;

	if (entry->type == DF_LOWER) {
		struct stat oldst;

		res = stat_ino(DF_LOWER, entry, &oldst);
		if (res < 0)
			return res;

		if (!copy_attrs(&oldst, st, to_set))
			return 0;

		res = copy_up(entry, &oldst);
		if (res < 0)
			return res;
	} else {
		res = setattr_ino(entry->type, entry, st, to_set);
		if (res < 0)
			return res;
	}

	return stat_ino(entry->type, entry, st);
}

static void delta_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
			  int to_set, struct fuse_file_info *fi)
{
	int res;

	(void) fi;

	res = do_setattr(ino, attr, to_set);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_attr(req, attr, attr_timeout);
}

static void delta_readlink(fuse_req_t req, fuse_ino_t ino)
{
	struct deltafs_entry *entry = get_entry(ino);
	char *link;
	int res;

	res = readlink_ino(entry->type, entry, &link);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_readlink(req, link);
	free(link);
}

static int entry_to_upper(struct deltafs_entry *entry)
{
	int res;

	if (entry->type == DF_LOWER) {
		struct stat parst;

		res = stat_ino(DF_LOWER, entry, &parst);
		if (res < 0)
			return res;

		res = copy_up(entry, &parst);
		if (res < 0)
			return res;
	}

	return 0;
}

static int create_parent(struct deltafs_entry *entry)
{
	return entry_to_upper(entry->parent);
}

static int do_newnode(struct deltafs_entry *entry, struct stat *st,
		      struct fuse_entry_param *e, const char *link)
{
	int res;
	int tmpfd = metafd;
	char *tmpfile = "tmp.create";
	int unlinkflag = 0;

	res = create_parent(entry);
	if (res < 0)
		return res;

	if (S_ISDIR(st->st_mode)) {
		unlinkflag |= AT_REMOVEDIR;
		res = mkdirat(tmpfd, tmpfile, st->st_mode);
		if (res == -1)
			res = -errno;
	} else if (S_ISREG(st->st_mode)) {
		res = openat(tmpfd, tmpfile, O_WRONLY | O_CREAT | O_EXCL,
			    st->st_mode);
		if (res == -1)
			res = -errno;
		else {
			close(res);
			res = 0;
		}
	} else if (S_ISLNK(st->st_mode)) {
		res = symlinkat(link, tmpfd, tmpfile);
		if (res == -1)
			res = -errno;
	} else {
		res = mknodat(tmpfd, tmpfile, st->st_mode, st->st_rdev);
		if (res == -1)
			res = -errno;
	}

	if (res == 0)
		res = set_owner(tmpfd, tmpfile, st);
	if (res == 0) {
		res = fstatat(tmpfd, tmpfile, &e->attr, AT_SYMLINK_NOFOLLOW);
		if (res == -1)
			res = -errno;
	}
	if (res == 0) {
		res = rename_to_ino(tmpfd, tmpfile, DF_UPPER, entry);
		if (res == -ENOTDIR && S_ISDIR(st->st_mode))
			res = rename_dir_to_whiteout(tmpfd, tmpfile, entry);
	}
	if (res == 0) {
		entry->type = DF_UPPER;
		entry->inode->nlookup++;

		e->ino = entry->inode->ino;
		e->attr.st_ino = e->ino;
		e->generation = 0;
		e->attr_timeout = attr_timeout;
		e->entry_timeout = entry_timeout;
	} else {
		unlinkat(tmpfd, tmpfile, unlinkflag);
	}

	return res;
}

static void delta_newnode(fuse_req_t req, fuse_ino_t parent, const char *name,
			  mode_t mode, dev_t rdev, const char *link)
{
	int res;
	struct fuse_entry_param e;
	struct deltafs_entry *entry;
	const struct fuse_ctx *ctx = fuse_req_ctx(req);
	struct stat st = {
		.st_mode = mode,
		.st_rdev = rdev,
		.st_uid = ctx->uid,
		.st_gid = ctx->gid,
	};

	/* fixme: make sure the entry doesn't exist before mknod */
	entry = find_entry(parent, name);
	if (!entry) {
		entry = new_entry(parent, name);
		if (!entry) {
			fuse_reply_err(req, ENOMEM);
			return;
		}
	}

	res = do_newnode(entry, &st, &e, link);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_entry(req, &e);
}

static void delta_mknod(fuse_req_t req, fuse_ino_t parent, const char *name,
			mode_t mode, dev_t rdev)
{
	delta_newnode(req, parent, name, mode, rdev, NULL);
}

static void delta_mkdir(fuse_req_t req, fuse_ino_t parent, const char *name,
			mode_t mode)
{
	delta_newnode(req, parent, name, (mode & MODEMASK) | S_IFDIR , 0, NULL);
}

static void delta_symlink(fuse_req_t req, const char *link, fuse_ino_t parent,
			  const char *name)
{
	delta_newnode(req, parent, name, S_IFLNK, 0, link);
}

static int create_whiteout_node(void)
{
	int res;
	struct stat st;

	if (whiteout_created)
		return 0;

	res = symlinkat("(WHITEOUT)", metafd, whiteout_name);
	if (res == -1)
		return -errno;

	res = fstatat(metafd, whiteout_name, &st, AT_SYMLINK_NOFOLLOW);
	if (res == -1)
		return -errno;

	whiteout_created = true;
	whiteout_ino = st.st_ino;

	return 0;
}

static int check_empty_merged(struct deltafs_entry *entry)
{
	int res;
	struct delta_dirent *list = NULL;
	struct delta_dirent *dent;

	res = merge_directories(entry, &list);
	if (res < 0)
		return res;

	for (dent = list; dent != NULL; dent = dent->next) {
		if (dent->ino != 0 && 
		    strcmp(dent->name, ".") != 0 &&
		    strcmp(dent->name, "..") != 0) {
			res = -ENOTEMPTY;
			break;
		}
	}

	free_dirlist(&list);
	return res;
}

static int check_empty(struct deltafs_entry *entry)
{
	int res;
	DIR *dp;
	struct dirent *de;
	bool is_empty = true;

	if (entry->type == DF_MERGE)
		return check_empty_merged(entry);

	res = opendir_ino(entry->type, entry, &dp);
	if (res < 0)
		return res;

	while ((de = readdir(dp)) != NULL) {
		if (strcmp(de->d_name, ".") != 0 &&
		    strcmp(de->d_name, "..") != 0) {
			is_empty = false;
			break;
		}
	}
	closedir(dp);

	return is_empty ? 0 : -ENOTEMPTY;
}

static int rename_to_deleted(struct deltafs_entry *entry)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(DF_DELETED, &lr, entry);
	if (res < 0)
		return res;

	/*
	 * This doesn't change st_ctime.  Not sure if this needs to be fixed,
	 * POSIX only requires st_ctime update if nlink is non-zero.
	 */
	res = rename_from_entry(entry, lr.fd, lr.path);
	lookup_finish(&lr);

	return res;
}

static int link_to_deleted(struct deltafs_entry *entry)
{
	struct lookup_result lr;
	int res;

	res = lookup_path(DF_UPPER, &lr, entry);
	if (res < 0)
		return res;

	res = link_to_ino(lr.fd, lr.path, DF_DELETED, entry);
	lookup_finish(&lr);

	return res;
}

static bool need_whiteout(struct deltafs_entry *entry)
{
	int res;
	struct stat dummy;

	if (entry->type != DF_UPPER)
		return true;

	if (entry->parent->type == DF_UPPER)
		return false;

	res = stat_ino(DF_LOWER, entry, &dummy);
	if (res == -ENOENT)
		return false;

	return true;
}

static int get_whiteout(struct lookup_result *wh)
{
	int res;

	res = create_whiteout_node();
	if (res < 0)
		return res;

	wh->closefd = false;
	wh->fd = metafd;
	snprintf(wh->path, sizeof(pathstring), "tmp.whiteout");

	res = linkat(metafd, whiteout_name, wh->fd, wh->path, 0);
	if (res == -1)
		return -errno;

	return 0;
}

static int do_remove(struct deltafs_entry *entry, bool is_dir)
{
	int res;
	struct lookup_result wh;
	bool linked = false;
	bool whiteout;

	if (is_dir) {
		res = check_empty(entry);
		if (res < 0)
			return res;
	}

	res = create_parent(entry);
	if (res < 0)
		return res;

	whiteout = need_whiteout(entry);
	if (whiteout) {
		res = get_whiteout(&wh);
		if (res < 0)
			return res;
	}

	if (!is_dir && entry_has_alias(entry)) {
		if (!whiteout)
			res = unlink_ino(entry->type, entry, 0);
		else
			res = rename_to_ino(wh.fd, wh.path, DF_UPPER, entry);
		if (res < 0)
			goto error;

		remove_from_name_hash(entry);
		remove_alias(entry);
		free_entry(entry);
		return 0;
	} else if (entry->type != DF_LOWER) {
		if (!is_dir && whiteout && entry->type == DF_UPPER) {
			res = link_to_deleted(entry);
			if (res < 0)
				goto error;
			linked = true;
		} else {
			res = rename_to_deleted(entry);
			if (res < 0)
				goto error;
		}
	}

	if (whiteout) {
		res = rename_to_ino(wh.fd, wh.path, DF_UPPER, entry);
		if (res < 0)
			goto error;
	}

	if (is_dir && opaque_created) {
		if (entry->type == DF_UPPER)
			unlink_ino(DF__MERGEDIR, entry, 0);
		else if (entry->type == DF_MERGE)
			unlink_ino(DF__MERGEDIR, entry, AT_REMOVEDIR);
	}

	remove_from_name_hash(entry);
	if (entry->type != DF_LOWER) {
		entry->parent = NULL;
		entry->type = DF_DELETED;
	}
	return 0;

error:
	if (linked)
		unlink_ino(DF_DELETED, entry, 0);
	if (whiteout)
		unlinkat(wh.fd, wh.path, 0);

	return res;
}

static void delta_unlink(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	int res;
	struct deltafs_entry *entry;

	entry = find_entry(parent, name);
	res = do_remove(entry, false);
	fuse_reply_err(req, -res);
}

static void delta_rmdir(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	int res;
	struct deltafs_entry *entry;

	entry = find_entry(parent, name);
	res = do_remove(entry, true);
	fuse_reply_err(req, -res);
}

static int do_rename(struct deltafs_entry *entry,
		     struct deltafs_entry *newentry)
{
	int res;
	struct stat st;
	struct lookup_result src;
	bool dst_exists = false;
	bool dst_is_dir = false;
	bool linked = false;
	struct lookup_result wh;
	bool whiteout;
	char *newname = NULL;
	struct deltafs_entry *newparent;

	if (newentry->type != DF_UNKNOWN) {
		res = stat_ino(newentry->type, newentry, &st);
		if (res == 0) {
			dst_exists = true;
			if (S_ISDIR(st.st_mode))
				dst_is_dir = true;
		} else if (res != -ENOENT) {
			return res;
		}
	}

	if (dst_exists && dst_is_dir) {
		res = check_empty(newentry);
		if (res < 0)
			return res;
	}

	if (entry->type != DF_UPPER) {
		res = stat_ino(entry->type, entry, &st);
		if (res < 0)
			return res;

		if (S_ISDIR(st.st_mode))
			return -EXDEV;

		res = copy_up(entry, &st);
		if (res < 0)
			return res;
	}

	res = create_parent(newentry);
	if (res < 0)
		return res;

	whiteout = need_whiteout(entry);
	if (whiteout) {
		res = get_whiteout(&wh);
		if (res < 0)
			return res;
	}
	res = lookup_path(DF_UPPER, &src, entry);
	if (res < 0)
		goto error;

	if (dst_exists && newentry->type != DF_LOWER &&
	    !entry_has_alias(newentry)) {
		if (!dst_is_dir && newentry->type == DF_UPPER) {
			res = link_to_deleted(newentry);
			if (res < 0)
				goto error;
			linked = true;
		} else {
			res = rename_to_deleted(newentry);
			if (res < 0)
				goto error;
		}
	}

	res = -ENOMEM;
	newname = strdup(newentry->name);
	if (newname == NULL)
		goto error;

	res = rename_to_ino(src.fd, src.path, DF_UPPER, newentry);
	if (res == -ENOTDIR)
		res = rename_dir_to_whiteout(src.fd, src.path, newentry);
	if (res < 0)
		goto error;

	if (whiteout) {
		res = rename_to_ino(wh.fd, wh.path, DF_UPPER, entry);
		if (res < 0)
			goto error;
	}

	if (opaque_created) {
		if (entry->type == DF_UPPER)
			unlink_ino(DF__MERGEDIR, entry, 0);
		else if (entry->type == DF_MERGE)
			unlink_ino(DF__MERGEDIR, entry, AT_REMOVEDIR);
	}

	newparent = newentry->parent;
	if (dst_exists) {
#if FUSE_VERSION >= 28
		/* kernel module forgot to invalidate target attributes */
		fuse_lowlevel_notify_inval_inode(fuse_ch, newentry->inode->ino,
						 -1, -1);
#endif
		remove_from_name_hash(newentry);
		if (entry_has_alias(newentry)) {
			remove_alias(newentry);
			free_entry(newentry);
		}
		if (newentry->type != DF_LOWER) {
			newentry->parent = NULL;
			newentry->type = DF_DELETED;
		}
	}
	remove_from_name_hash(entry);
	free(entry->name);
	entry->name = newname;
	entry->parent = newparent;
	add_to_name_hash(entry);

	return 0;

error:
	free(newname);
	if (linked)
		unlink_ino(DF_DELETED, entry, 0);
	if (whiteout)
		unlinkat(wh.fd, wh.path, 0);

	return res;
}

static void delta_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
			 fuse_ino_t newparent, const char *newname)
{
	int res;
	struct deltafs_entry *entry;
	struct deltafs_entry *newentry;
	bool created = false;

	entry = find_entry(parent, name);
	newentry = find_entry(newparent, newname);
	if (!newentry) {
		struct deltafs_entry *newdir = get_entry(newparent);

		newentry = create_entry(newdir, newname, NULL, DF_UNKNOWN);
		if (newentry == NULL) {
			fuse_reply_err(req, ENOMEM);
			return;
		}
		created = true;
	}
	res = do_rename(entry, newentry);
	if (created)
		free_entry(newentry);
	fuse_reply_err(req, -res);
}

static int do_link(struct deltafs_entry *entry, struct deltafs_entry *newentry,
		   struct fuse_entry_param *e)
{
	int tmpfd = metafd;
	char *tmpfile = "tmp.link";
	int res;

	if (entry->type != DF_UPPER) {
		struct stat st;

		res = stat_ino(entry->type, entry, &st);
		if (res < 0)
			return res;

		res = copy_up(entry, &st);
		if (res < 0)
			return res;
	}

	res = link_from_ino(entry->type, entry, tmpfd, tmpfile);
	if (res < 0)
		return res;

	res = fstatat(tmpfd, tmpfile, &e->attr, AT_SYMLINK_NOFOLLOW);
	if (res == -1) {
		res = -errno;
		goto cleanup_tmp;
	}

	res = rename_to_ino(tmpfd, tmpfile, DF_UPPER, newentry);
	if (res < 0)
		goto cleanup_tmp;

	add_to_name_hash(newentry);
	newentry->inode->nlookup++;

	e->ino = newentry->inode->ino;
	e->attr.st_ino = e->ino;
	e->generation = 0;
	e->attr_timeout = attr_timeout;
	e->entry_timeout = entry_timeout;

	return 0;

cleanup_tmp:
	unlinkat(tmpfd, tmpfile, 0);
	return res;
}

static void delta_link(fuse_req_t req, fuse_ino_t ino, fuse_ino_t newparent,
		const char *newname)
{
	int res;
	struct deltafs_entry *entry;
	struct deltafs_entry *newdir;
	struct deltafs_entry *newentry;
	struct fuse_entry_param e;

	entry = get_entry(ino);
	newdir = get_entry(newparent);

	newentry = create_entry(newdir, newname, entry->inode, DF_UPPER);
	if (!newentry) {
		fuse_reply_err(req, ENOMEM);
		return;
	}

	res = do_link(entry, newentry, &e);
	if (res < 0) {
		free_entry(newentry);
		fuse_reply_err(req, -res);
	} else {
		fuse_reply_entry(req, &e);
	}
}

static inline struct delta_dirp *get_dirp(struct fuse_file_info *fi)
{
	return (struct delta_dirp *) (uintptr_t) fi->fh;
}

static void free_dirp(struct delta_dirp *d)
{
	free_dirlist(&d->first);
	free(d);
}

static void delta_opendir(fuse_req_t req, fuse_ino_t ino,
			  struct fuse_file_info *fi)
{
	struct delta_dirp *d;
	int res;

	d = malloc(sizeof(struct delta_dirp));
	if (d == NULL) {
		fuse_reply_err(req, ENOMEM);
		return;
	}

	res = do_opendir(ino, d);
	if (res < 0) {
		free_dirp(d);
		fuse_reply_err(req, -res);
	} else {
		fi->fh = (unsigned long) d;
		fuse_reply_open(req, fi);
	}
}

static void delta_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
			     off_t off, struct fuse_file_info *fi)
{
	struct delta_dirp *d = get_dirp(fi);
	char *buf;
	int res;

	buf = malloc(size);
	if (buf == NULL) {
		fuse_reply_err(req, ENOMEM);
		return;
	}

	res = do_readdir(req, ino, buf, size, off, d);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_buf(req, buf, res);
	free(buf);
}

static void delta_releasedir(fuse_req_t req, fuse_ino_t ino,
			     struct fuse_file_info *fi)
{
	struct delta_dirp *d = get_dirp(fi);

	(void) ino;

	closedir(d->dp);
	free_dirp(d);
	fuse_reply_err(req, 0);
}

static int do_open(fuse_ino_t ino, struct fuse_file_info *fi)
{
	struct deltafs_entry *entry = get_entry(ino);
	int fd;
	int flags = (fi->flags & O_ACCMODE) | O_NOFOLLOW;
	int res;

	if (entry->type == DF_LOWER) {
		if ((fi->flags & O_ACCMODE) != O_RDONLY) {
			struct stat st;

			res = stat_ino(DF_LOWER, entry, &st);
			if (res < 0)
				return res;

			res = copy_up(entry, &st);
			if (res < 0)
				return res;
		} else if (deltafs.o_direct) {
			flags |= O_DIRECT;
		}
	}

	fd = open_ino(entry->type, entry, flags);
	if (fd < 0)
		return fd;

	fi->fh = fd;
	fi->keep_cache = 1;

	return 0;
}

static void delta_open(fuse_req_t req, fuse_ino_t ino,
			struct fuse_file_info *fi)
{
	int res = do_open(ino, fi);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_open(req, fi);
}

static void delta_read(fuse_req_t req, fuse_ino_t ino, size_t size,
		       off_t off, struct fuse_file_info *fi)
{
	int res;
	void *buf;

	(void) ino;

	res = posix_memalign(&buf, pagesize, size);
	if (res != 0) {
		fuse_reply_err(req, res);
		return;
	}

	res = pread(fi->fh, buf, size, off);
	if (res == -1)
		fuse_reply_err(req, errno);
	else
		fuse_reply_buf(req, buf, res);
	free(buf);
}

static void delta_write(fuse_req_t req, fuse_ino_t ino, const char *buf,
			size_t size, off_t off, struct fuse_file_info *fi)
{
	int res;

	(void) ino;

	res = pwrite(fi->fh, buf, size, off);
	if (res == -1)
		fuse_reply_err(req, errno);
	else
		fuse_reply_write(req, res);
}

static void delta_release(fuse_req_t req, fuse_ino_t ino,
			  struct fuse_file_info *fi)
{
	(void) ino;

	close(fi->fh);
	fuse_reply_err(req, 0);
}

static void delta_fsync(fuse_req_t req, fuse_ino_t ino, int datasync,
			struct fuse_file_info *fi)
{
	int res;

	(void) ino;

	if (datasync)
		res = fdatasync(fi->fh);
	else
		res = fsync(fi->fh);
	if (res == -1)
		fuse_reply_err(req, errno);
	else
		fuse_reply_err(req, 0);
}

static int do_statfs(struct statvfs *st)
{
	int res;

	res = fstatvfs(upperfd, st);
	if (res < 0)
		return -errno;

	return 0;
}

static void delta_statfs(fuse_req_t req, fuse_ino_t ino)
{
	struct statvfs st;
	int res;

	(void) ino;

	res = do_statfs(&st);
	if (res < 0)
		fuse_reply_err(req, -res);
	else
		fuse_reply_statfs(req, &st);
}

static struct fuse_lowlevel_ops delta_oper = {
	.lookup		= delta_lookup,
	.forget		= delta_forget,
	.getattr	= delta_getattr,
	.setattr	= delta_setattr,
	.readlink	= delta_readlink,
	.mknod		= delta_mknod,
	.mkdir		= delta_mkdir,
	.unlink		= delta_unlink,
	.rmdir		= delta_rmdir,
	.symlink	= delta_symlink,
	.rename		= delta_rename,
	.link		= delta_link,
	.open		= delta_open,
	.read		= delta_read,
	.write		= delta_write,
	.release	= delta_release,
	.fsync		= delta_fsync,
	.opendir	= delta_opendir,
	.readdir	= delta_readdir,
	.releasedir	= delta_releasedir,
	.statfs		= delta_statfs,
};

enum  {
	KEY_HELP,
	KEY_VERSION,
};

#define DELTAFS_OPT(t, p, v) { t, offsetof(struct deltafs_config, p), v }

static const struct fuse_opt delta_opts[] = {
	DELTAFS_OPT("lowerdir=%s",	lowerdir, 0),
	DELTAFS_OPT("upperdir=%s",	upperdir, 0),
	DELTAFS_OPT("o_direct=yes",	o_direct, 1),
	DELTAFS_OPT("o_direct=no",	o_direct, 0),
	FUSE_OPT_KEY("-h",		KEY_HELP),
	FUSE_OPT_KEY("--help",		KEY_HELP),
	FUSE_OPT_KEY("-V",		KEY_VERSION),
	FUSE_OPT_KEY("--version",	KEY_VERSION),
	FUSE_OPT_END
};

static void usage(const char *progname, FILE *f)
{
	fprintf(f, "usage: %s [opts] -o lowerdir=DIR,upperdir=DIR mountpoint\n",
		progname);
}

static int delta_opt_proc(void *data, const char *arg, int key,
			  struct fuse_args *outargs)
{
	(void) arg;
	(void) data;

	switch (key) {
	case KEY_HELP:
		usage(outargs->argv[0], stdout);
		printf(
		"General options:\n"
		"    -o opt,[opt...]        mount options\n"
		"    -h   --help            print help\n"
		"    -V   --version         print version\n"
		"    -d   -o debug          enable debug output (implies -f)\n"
		"    -f                     foreground operation\n"
		"\n"
		"Deltafs options:\n"
		"    -o lowerdir=DIR        read-only base layer\n"
		"    -o upperdir=DIR        read-write layer, stores changes\n"
		"    -o o_direct=yes        enable O_DIRECT on lower layer\n"
		"    -o o_direct=no         disable O_DIRECT on lower layer (default)\n"
		"\n"
		"FUSE options:\n");
		fflush(stdout);
		dup2(1, 2);
		fuse_opt_add_arg(outargs, "--help");
		fuse_mount(NULL, outargs);
		fuse_lowlevel_new(outargs, NULL, 0, NULL);
		exit(0);

	case KEY_VERSION:
		printf("deltafs version %s\n", PACKAGE_VERSION);
		fflush(stdout);
		dup2(1, 2);
		fuse_opt_add_arg(outargs, "--version");
		fuse_parse_cmdline(outargs, NULL, NULL, NULL);
		fuse_lowlevel_new(outargs, NULL, 0, NULL);
		exit(0);
	}
	return 1;
}

static void cleanup_one(const char *name)
{
	int fd;

	fd = openat(metafd, name, O_RDONLY | O_DIRECTORY | O_NOFOLLOW);
	if (fd == 0) {
		DIR *d = fdopendir(fd);
		if (d != NULL) {
			struct dirent *de;

			while ((de = readdir(d)) != NULL) {
				if (strcmp(de->d_name, ".") == 0 ||
				    strcmp(de->d_name, "..") == 0)
					continue;

				unlinkat(fd, de->d_name, 0);
			}
			closedir(d);
		} else {
			close(fd);
		}
	}
	unlinkat(metafd, name, AT_REMOVEDIR);
	unlinkat(metafd, name, 0);

}

static void cleanup_tmp(void)
{
	cleanup_one("tmp.copy");
	cleanup_one("tmp.create");
	unlinkat(metafd, "tmp.whiteout", 0);
	unlinkat(metafd, "tmp.link", 0);
}

int main(int argc, char *argv[])
{
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
	struct fuse_chan *ch;
	struct deltafs_inode *root_inode;
	struct deltafs_entry *root;
	char *mountpoint;
	struct stat st;
	int res;
	int foreground = 0;

	umask(0);
	pagesize = getpagesize();

	res = fuse_opt_parse(&args, &deltafs, delta_opts, delta_opt_proc);
	if (res == -1)
		exit(1);
	res = fuse_opt_insert_arg(&args, 1, "-oallow_other,default_permissions,nonempty,big_writes,dev,suid");
	if (res == -1)
		exit(1);

	if (deltafs.lowerdir == NULL || deltafs.upperdir == NULL) {
		usage(argv[0], stderr);
		exit(1);
	}

	lowerfd = open(deltafs.lowerdir, O_RDONLY | O_DIRECTORY | O_NOFOLLOW);
	if (lowerfd == -1)
		err(1, "failed to open %s", deltafs.lowerdir);

	upperfd = open(deltafs.upperdir, O_RDONLY | O_DIRECTORY | O_NOFOLLOW);
	if (upperfd == -1)
		err(1, "failed to open %s", deltafs.upperdir);

	metafd = upperfd;

	cleanup_tmp();

	res = fstatat(metafd, whiteout_name, &st, AT_SYMLINK_NOFOLLOW);
	if (res == -1) {
		if (errno != ENOENT)
			err(1, "failed to stat %s/%s",
			    deltafs.upperdir, whiteout_name);
	} else {
		whiteout_ino = st.st_ino;
		whiteout_created = true;
	}

	res = fstatat(metafd, opaque_name, &st, AT_SYMLINK_NOFOLLOW);
	if (res == -1) {
		if (errno != ENOENT)
			err(1, "failed to stat %s/%s",
			    deltafs.upperdir, opaque_name);
	} else {
		opaque_ino = st.st_ino;
		opaque_created = true;
	}

	id_table_size = 32768;
	name_table_size = 32768;

	id_table = calloc(id_table_size, sizeof(struct deltafs_entry *));
	name_table = calloc(name_table_size, sizeof(struct deltafs_entry *));
	if (id_table == NULL || name_table == NULL)
		errx(1, "memory allocation failed");

	root_inode = create_inode(FUSE_ROOT_ID);
	root = create_entry(NULL, ".", root_inode, DF_UNKNOWN);
	if (root == NULL || root_inode == NULL)
		errx(1, "memory allocation failed");
	add_to_id_hash(root_inode);
	res = check_type(root, &root->type, &st);
	if (res < 0)
		errx(1, "failed to determine root type: %s", strerror(-res));

	res = -1;
	if (fuse_parse_cmdline(&args, &mountpoint, NULL, &foreground) != -1 &&
	    (ch = fuse_mount(mountpoint, &args)) != NULL) {
		struct fuse_session *se;

		se = fuse_lowlevel_new(&args, &delta_oper,
				       sizeof(delta_oper), NULL);
		if (se != NULL) {
			res = fuse_daemonize(foreground);
			if (res == 0) {
				if (fuse_set_signal_handlers(se) != -1) {
					fuse_session_add_chan(se, ch);
					fuse_ch = ch;
					res = fuse_session_loop(se);
					fuse_remove_signal_handlers(se);
					fuse_session_remove_chan(ch);
				}
			}
			fuse_session_destroy(se);
		}
		fuse_unmount(mountpoint, ch);
	}
	fuse_opt_free_args(&args);

	return res ? 1 : 0;
}
